// -*- mode: c; coding: utf-8 -*- */
//
// Copyright 2010, 2011, Matthias Andreas Benkard.
// Modifications for lazy eval: Copyright 2011, Christoph-Simon Senjak
//
//-----------------------------------------------------------------------------
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//

// An implementation of a bitmapped Patricia tree.

#ifndef __BITMAPPED_PATRICIA_TREE_H
#define __BITMAPPED_PATRICIA_TREE_H

#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>

#define BPT_ENABLE_DEALLOC_HOOKS 1

/* these values are normally set by the configure script of mulklib */

typedef intptr_t bpt_key_t;
typedef intptr_t bpt_key_bitmask_t;

#define CHUNK_LENGTH ((int)log2(sizeof(bpt_key_bitmask_t)*8))
#define KEY_LENGTH (sizeof(bpt_key_t)*8)
#define OFFSET_MASK ((1 << CHUNK_LENGTH) - 1)
#define MAX_CHUNKS (KEY_LENGTH / CHUNK_LENGTH + ((KEY_LENGTH % CHUNK_LENGTH == 0) ? 0 : 1))
#define LAST_CHUNK_LENGTH (KEY_LENGTH - ((MAX_CHUNKS - 1) * CHUNK_LENGTH))

enum bpt_tag {
  BPT_LEAF,
  BPT_INNER_NODE
};

struct bpt;
typedef struct bpt *bpt_t;

void *bpt_get(bpt_t bpt, bpt_key_t key);
bool bpt_has_key(bpt_t bpt, bpt_key_t key);
void **bpt_get_pointer(bpt_t bpt, bpt_key_t key);
bpt_t bpt_assoc(bpt_t bpt, bpt_key_t key, void *item);
bpt_t bpt_dissoc(bpt_t bpt, bpt_key_t key);
void init_bpt_leaf(bpt_t leaf, bpt_key_t key, void *value);
bpt_t bpt_make_leaf(bpt_key_t key, void *value);
void bpt_retain(bpt_t bpt);
void bpt_release(bpt_t bpt);
void bpt_dealloc(bpt_t bpt);
void bpt_seal(bpt_t bpt);

#ifdef BPT_ENABLE_DEALLOC_HOOKS
void bpt_set_dealloc_hook(bpt_t bpt, bpt_key_t key, void (*hook)(bpt_key_t key, void* value));
#endif
#endif
